using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.IO;
using ArincAPI;
using ArincAPI.DataTypes;
using System.Runtime.InteropServices;

namespace ArincCsApp
{
   public static class ArincFunctions
   {
      public static Form1 mainForm;
      public static Int16 gTxChannel, gRxChannel;  // local selected transmit and receive channels
      public static Int16 gInternalExternalWrap;   // local selected receiver wrap mode
      public static short deviceNumber;            // local board number
      public static Thread monitorThread = null;   // handle for the receiver monitor thread
      public static bool sessionOpened = false;    // local flag indicating a session was successfully opened with the board
      public static Int16 fwVersion = 0;           // Board Firmware Version
      public static Int32 fwVersionMajor = 0;      // Board Firmware Version Major 
      public static Int32 fwVersionMinor = 0;      // Board Firmware Version Minor
      public static string filePath = @"ArincCsApp.log";  // log file
      public static ArincAPI.DataTypes.AR_SCHEDULED_MSG_ENTRY_TYPE message = new AR_SCHEDULED_MSG_ENTRY_TYPE(); // scheduled message entry

      /////////////////////////////////////////////////////////////////////////////////
      // these delegates and the next 2 routines provide thread-safe access to the
      // text box tbOutput and optional message logging to "ArincCsApp.log"
      /////////////////////////////////////////////////////////////////////////////////
      public delegate void AddToOutputDelegate(String s);
      public static AddToOutputDelegate addDelegate;
            
      public static void AddStringToOutput(String s)
      {
         mainForm.tbOutput.AppendText(s);
      }

      public static void OutputMessage(String s)
      {
         if (mainForm.tbOutput.InvokeRequired)
         {
            mainForm.tbOutput.Invoke(addDelegate, new object[] { s });
         }
         else
         {
            mainForm.tbOutput.AppendText(s);
         }
         if (mainForm.cbLogMessages.Checked)
         {
            File.AppendAllText(filePath, s);
         }
      }
      /////////////////////////////////////////////////////////////////////////////////
      // end text output routines
      /////////////////////////////////////////////////////////////////////////////////

      // constructor
      static ArincFunctions()
      {
         addDelegate = new AddToOutputDelegate(AddStringToOutput);
      }

      /////////////////////////////////////////////////////////////////////////////////
      // routine invoked as a new thread to monitor the specified receive channel for 
      // messages via the ar_getnext_xt routine (having a built-in delay of 1/2 second)
      /////////////////////////////////////////////////////////////////////////////////
      public static void Monitor_Reception()
      {
         short status = ArincAPI.Constants.ARS_GOTDATA;
         UInt32 rxMessage = 0;
         ArincAPI.DataTypes.AR_TIMETAG_TYPE timeStampStruct = new AR_TIMETAG_TYPE();
         
         // assure we should be entering the monitor-receiver-forever loop
         if (mainForm.monitorOnly.Checked || (message.rate != 0))
         {
            // until terminated, this thread continuously monitors the specified receiver
            while (true) {
               status = ArincAPI.API.Get_Next_XT(deviceNumber, gRxChannel, ref rxMessage, timeStampStruct);
               if (status == ArincAPI.Constants.ARS_GOTDATA)
               {
                  OutputMessage(String.Format("Rx Channel {0} Received: " + rxMessage.ToString("X8") + "  Time Stamp: {1} microseconds" + Environment.NewLine, gRxChannel, timeStampStruct.timeTag));
               }
               else
               {
                  OutputMessage(String.Format("Rx Channel {0} Reception Status: {1}" + Environment.NewLine, gRxChannel, status));
               }
            }
         }
      }

      /////////////////////////////////////////////////////////////////////////////////////
      // routine invoked when the Open Button button is invoked, performing the following:
      //   setup the API for multi-threaded operation
      //   open a session with the card
      //   read the firmware version installed on the card
      /////////////////////////////////////////////////////////////////////////////////////
      public static short Init(short device)
      {
         short status;

         // configure the API for multi-thread protected access to the board, since most GUI applications are multi-threaded
         status = ArincAPI.API.Set_Preload_Config(device, Constants.ARU_CONCURRENCY_MODE, Constants.AR_CONC_MULTITHRD);

         // open a session with the board
         status = ArincAPI.API.Open_Session(device);

         if (status == Constants.ARS_NORMAL)
         {
            sessionOpened = true; // set this flag as an indication the session requires closure when stopped

            // request and display the firmware version installed on the board
            status = ArincAPI.API.Get_Device_Config(device, 0, Constants.ARU_FW_VERSION, ref fwVersion);
            fwVersionMajor = (fwVersion & 0x00F0) >> 4;
            fwVersionMinor = (fwVersion & 0x000F);
            mainForm.tbFwVersion.AppendText(fwVersionMajor.ToString("X1") + "." + fwVersionMinor.ToString("X1"));
         }
         // check for a successful session open and firmware version acquisition
         if (status == Constants.ARS_NORMAL)
         {
            deviceNumber = device;  // make the device number visible to the receiver monitor thread
            OutputMessage(String.Format("Successfully opened a session with Device {0}" + Environment.NewLine, device));
            OutputMessage(String.Format("Detected {0} transmit channels and {1} receive channels" + Environment.NewLine, ArincAPI.API.Num_TX_Chans(device), ArincAPI.API.Num_RX_Chans(device)));
         }
         else
         {
            OutputMessage(String.Format("Failed to open a session with Device {0}" + Environment.NewLine + "{1}", device, ArincAPI.API.Get_Error(status)));
         }
         return status;
      }

      /////////////////////////////////////////////////////////////////////////////////
      // routine invoked when the Go button is invoked, performing the following:
      //   configure the specified transmit and receive channels based on the
      //     radio buttons and combo box selections
      //   for internal or external wrap, defined a scheduled message to transmit
      //   reset the board time to zero
      //   enable all message processing on the board
      //   create the receiver monitor thread (above)
      /////////////////////////////////////////////////////////////////////////////////
      public static short Activate(short device)
      {
         short status = Constants.ARS_NORMAL;
         short txBusSpeed;
         short rxBusSpeed;
         short txParity;
         short rxParity;

         // initialize the schedueld message index used as a flag during app termination to invoke message transmit disable
         message.messageIndex = 1024;

         // check for valid channel selections
         gTxChannel = (short)mainForm.cbTxChannel.SelectedIndex;
         gRxChannel = (short)mainForm.cbRxChannel.SelectedIndex;

         if (gTxChannel >= ArincAPI.API.Num_TX_Chans(device))
         {
            OutputMessage(String.Format("The specified Tx Channel {0} exceeds the ARINC 429 transmit channel count available on this Device: {1}" + Environment.NewLine, gTxChannel, ArincAPI.API.Num_TX_Chans(device)));
            status = Constants.ARS_INVHARVAL;
         }
         else if (gRxChannel >= ArincAPI.API.Num_RX_Chans(device))
         {
            OutputMessage(String.Format("The specified Rx Channel {0} exceeds the ARINC 429 receive channel count available on this Device: {1}" + Environment.NewLine, gRxChannel, ArincAPI.API.Num_RX_Chans(device)));
            status = Constants.ARS_INVHARVAL;
         }

         // if the session was opened successfully and the channel selections are valid, configure the channels for the selection operation
         if (status == Constants.ARS_NORMAL)
         {
            // setup the local receiver internal wrap, and both transmit and receive speed and parity selections
            gInternalExternalWrap = (short)((mainForm.monitorOnly.Checked || mainForm.wrapExtBus.Checked) ? Constants.AR_WRAP_OFF : Constants.AR_WRAP_ON);
            txBusSpeed = mainForm.cbTxBusSpeed.SelectedIndex == 0 ? Constants.ARU_SPEED_HIGH : Constants.ARU_SPEED_LOW;
            rxBusSpeed = mainForm.cbRxBusSpeed.SelectedIndex == 0 ? Constants.ARU_SPEED_HIGH : Constants.ARU_SPEED_LOW;
            rxParity = mainForm.cbRxParity.SelectedIndex == 0 ? Constants.AR_ON : Constants.AR_OFF;
            switch (mainForm.cbTxParity.SelectedIndex)
            {
               case 0:
                  txParity = Constants.ARU_PARITY_ODD;
                  break;
               case 1:
                  txParity = Constants.ARU_PARITY_EVEN;
                  break;
               default:
               case 2:
                  txParity = Constants.ARU_PARITY_NONE;
                  break;
            }
            
            // assign all of the channel attributes using Set_Device_Config
            status = ArincAPI.API.Set_Device_Config(device, gRxChannel, Constants.ARU_RECV_MODE, gInternalExternalWrap);
            if (status != Constants.ARS_NORMAL)
            {
               OutputMessage(String.Format("Set_Device_Config(ARU_RECV_MODE) Failed for Device {0} Receive Channel {1}" + Environment.NewLine, device, gRxChannel));
            }
            else
            {
               status = ArincAPI.API.Set_Device_Config(device, gRxChannel, Constants.ARU_RX_BITRATE, rxBusSpeed);
               if (status != Constants.ARS_NORMAL)
               {
                  OutputMessage(String.Format("Set_Device_Config(ARU_RX_BITRATE) Failed for Device {0} Receive Channel {1}" + Environment.NewLine, device, gRxChannel));
               }
               else
               {
                  status = ArincAPI.API.Set_Device_Config(device, gRxChannel, Constants.ARU_RX_PARITY, rxParity);
                  if (status != Constants.ARS_NORMAL)
                  {
                     OutputMessage(String.Format("Set_Device_Config(ARU_RX_PARITY) Failed for Device {0} Receive Channel {1}" + Environment.NewLine, device, gRxChannel));
                  }
                  else
                  {
                     status = ArincAPI.API.Set_Device_Config(device, gTxChannel, Constants.ARU_TX_BITRATE, txBusSpeed);
                     if (status != Constants.ARS_NORMAL)
                     {
                        OutputMessage(String.Format("Set_Device_Config(ARU_TX_BITRATE) Failed for Device {0} Transmit Channel {1}" + Environment.NewLine, device, gTxChannel));
                     }
                     else
                     {
                        status = ArincAPI.API.Set_Device_Config(device, gTxChannel, Constants.ARU_TX_PARITY, txParity);
                        if (status != Constants.ARS_NORMAL)
                        {
                           OutputMessage(String.Format("Set_Device_Config(ARU_TX_PARITY) Failed for Device {0} Transmit Channel {1}" + Environment.NewLine, device, gTxChannel));
                        }
                     }
                  }
               }
            }

            // disable the transmitter if we're not performing external wrap
            if ((status == Constants.ARS_NORMAL) && (mainForm.wrapExtBus.Checked == false))
            {
               status = ArincAPI.API.Set_Device_Config(device, gTxChannel, Constants.ARU_TX_DISABLE, Constants.AR_ON);
               if (status != Constants.ARS_NORMAL)
               {
                  OutputMessage(String.Format("Set_Device_Config(ARU_TX_DISABLE) Failed for Device {0} Transmit Channel {1}" + Environment.NewLine, device, gTxChannel));
               }
            }

            // create a single scheduled (periodic) message having the value 0xdeadbeef
            if ((status == Constants.ARS_NORMAL) && (mainForm.wrapExtBus.Checked || mainForm.wrapIntBus.Checked))
            {
               message.board = (UInt32)device;
               message.channel = (UInt32)gTxChannel;
               message.rate = (UInt32)mainForm.numTransmitMsgRate.Value;
               message.start = 0;
               message.txCount = 0xFFFFFFFF;
               message.data = (UInt32)mainForm.numTransmitMessage.Value;

               if (message.messageIndex < 1024)
               {
                   status = ArincAPI.API.Modify_Msg_Block(1, message);
               }
               else
               {
                   status = ArincAPI.API.Define_Msg_Block(1, message);
               }
            }

            // reset the onboard timer so received message time-stamps start around time 0+
            if (status == Constants.ARS_NORMAL)
            {
               ArincAPI.DataTypes.AR_TIMETAG_TYPE timeSetStruct = new AR_TIMETAG_TYPE();
               timeSetStruct.timeTag = (UInt64)0;
               timeSetStruct.timeTagFormat = (UInt64)Constants.AR_TIMETAG_INT_USEC_64BIT;
               status = ArincAPI.API.Set_Time(device, timeSetStruct);

               // enable all message processing on the board
               status = ArincAPI.API.Go(device);

               // create the receiver monitor thread
               monitorThread = new Thread(new ThreadStart(ArincFunctions.Monitor_Reception));
               monitorThread.Start();
            }
         }

         return status;
      }

      /////////////////////////////////////////////////////////////////////////////////
      // routine invoked when the Stop button is invoked, performing the following:
      //   if a scheduled message was defined for transmission, disable it
      //   if the receiver monitor thread is active, terminate it and wait for closure
      //   disable message processing and close the session with the card
      /////////////////////////////////////////////////////////////////////////////////
      public static short Terminate(short device)
      {
         short status = Constants.ARS_NORMAL;

         // if a scheduled message was created (index reset between 0 and 1023), disable
         // the message via assignment of a message rate value of 0 milliseconds
         if (message.messageIndex < 1024)
         {
            message.rate = 0;
            status = ArincAPI.API.Modify_Msg_Block(1, message);
         }

         // if the receiver monitor thread is active, terminate it and wait for closure
         if ((monitorThread != null) && (monitorThread.IsAlive))
         {
            monitorThread.Abort();
            monitorThread.Join();
         }

         // disable message processing
         status = ArincAPI.API.Stop(device);
         return status;
      }

   
      ///////////////////////////////////////////////////////////////////////////////////////
      // routine invoked when the Close Session button is invoked, performing the following:
      //   check for invocation while "Go/Stop" is "Go" and if Rx monitor thread is running,
      //     terminate it...
      //   close the session with the card
      ///////////////////////////////////////////////////////////////////////////////////////
      public static short Close(short device)
      {
         short status;
         // if the Close Session button is invoked while actively monitoring a receiver, terminate that thread first
         if ((monitorThread != null) && (monitorThread.IsAlive))
         {
            status = Terminate(device);
         }
         // close the session
         status = ArincAPI.API.Close_Session(device);
         if (status != Constants.ARS_NORMAL)
         {
            OutputMessage(String.Format("Failed to close the session with Device {0}" + Environment.NewLine + "{1}", device, ArincAPI.API.Get_Error(status)));
         }
         else
         {
            OutputMessage(String.Format("Successfully closed the session with Device {0}" + Environment.NewLine, device));
         }
         sessionOpened = false;
         return status;
      }
   }
}
