/* #define AUTO_EMBEDDED_UPDATE */
/* #define DEFAULT_ISR */

/*====================================================================================================================*
 * FILE:                                   T S T _ C N F G . C
 *====================================================================================================================*
 *
 *   COPYRIGHT 2005 - 2016 ABACO SYSTEMS, INC.  ALL RIGHTS RESERVED.
 *  
 *   THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED
 *   ONLY IN ACCORDANCE WITH THE TERMS OF SUCH LICENSE AND WITH THE INCLUSION
 *   OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY OTHER COPIES THEREOF
 *   MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY OTHER PERSON.
 *   NO TITLE TO AND OWNERSHIP OF THE SOFTWARE IS HEREBY TRANSFERRED.
 *  
 *   THE INFORMATION IN THIS SOFTWARE IS SUBJECT TO CHANGE WITHOUT NOTICE
 *   AND SHOULD NOT BE CONSTRUED AS A COMMITMENT BY ABACO SYSTEMS, INC. 
 *
 *====================================================================================================================*
 *
 * FUNCTION:    Demonstration program for the CEI-x30 ARINC products.
 *
 * ENTRY POINTS:
 *              main/wrap                 - main program
 *              custom_interrupt_handler  - demonstartion of a custom isr
 *              test_basic_arinc_429      - test message wrap on ARINC 429 channels
 *              test_arinc_717            - test frame wrap on an ARINC 717 channel
 *              demo_advanced_arinc_429   - demonstrate filtering and scheduling
 *              demo_discrete_io_features - demonstrate use of Discrete I/O
 *              demo_irig_features        - demonstrate IRIG and timer API features
 *              demo_pci_interrupts       - demonstrate setup of PCI interrupts
 *              error_exit                - print a message and error exit.
 *
 * DESCRIPTION: The routines in this file are basic example test programs 
 *              transmitting data on a transmit channel(s) and retrieving it
 *              from the respective internally wrapped receive channel(s).
 *
 *--------------------------------------------------------------------------------------------------------------------*
 *          
 * HISTORY:   
 *
 *   Date   Initials      Comments
 * -------- --------      ------------------------------------------------------
 * 03/16/04   BDW         Created.
 * 09/16/05   BDW         Added support for Linux x86.
 * 11/29/05   BDW         Added support for Integrity.
 * 08/24/06   BDW         Incorporated CEI_TYPES.H data type definitions in 
 *                          support of 64-bit Linux operating systems - this 
 *                          modified every data declaration and prototype.
 *                        Added a demonstration of Discrete I/O usage
 * 05/11/07   BDW         Added support for the CEI-530
 * 08/22/07   BDW         Added support for the R830RX
 * 12/06/07   BDW         Added support for the RAR-EC and RAR-CPCI
 * 03/31/08   BDW         Added support for Linux command line parameters A, I,
 *                        D & E, to optionally execute Advanced, IRIG or Discrete
 *                        I/O tests and disable internal wrap.
 * 06/23/09   BDW         Added a test for the Enhanced CEI-x30 Interface and
 *                        invocation to update any standard device to enhanced.
 *                        Provided the compiler define AUTO_EMBEDDED_UPDATE to
 *                        allow embedded builds the ability to auto-update devices.
 * 07/08/09   BDW         Added a demonstration of the hardware interrupt feature.
 *
 *                  Ver.  Comments
 *                  ----  --------------------------------------------------------------------------------------------
 * 01/08/15   BDW   4.60  Added detection and indication of s/w programmable board cfg
 * 03/08/16   BDW   4.90  Added detection and indication of API and firmware versions
 * 10/18/16   BDW   4.90  Introduced use of CEIx30 API-specific type defines via the file CEIX30_TYPES.H allowing
 *                        for easy build to common use of only signed integer return types and unsigned integer
 *                        parameter types (with the exception of 64-bit types); specifically replacing use of
 *                        CEI_INT16 with parameter specific types for functions using short integer parameters.
 *
 *====================================================================================================================*/ 

/*---------------------------------------------------------------------------*
 *          INCLUDES, DEFINES and GLOBALS
 *---------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "cdev_api.h"  
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
#include <conio.h>
#endif


/*===========================================================================*
 *                           P R O T O T Y P E S
 *===========================================================================*/
CDEV_API_RET_TYPE demo_advanced_arinc_429 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount );
CDEV_API_RET_TYPE test_arinc_717 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount );
CDEV_API_RET_TYPE test_basic_arinc_429 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount );
CDEV_API_RET_TYPE demo_discrete_io_features ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount );
CDEV_API_RET_TYPE demo_irig_features ( CDEV_BOARD_TYPE board );
CDEV_API_RET_TYPE demo_pci_interrupts ( CDEV_BOARD_TYPE board );
CEI_VOID demo_programmable_channels ( CDEV_BOARD_TYPE board );
CEI_VOID check_enhanced_firmware_capability (CDEV_BOARD_TYPE board, CEI_NATIVE_INT restoreConfig);
CEI_VOID error_exit ( CDEV_PARM_SSI_TYPE status );

/*===========================================================================*
 *              G L O B A L   C O N S T A N T S   A N D   D A T A
 *===========================================================================*/
#define N_WORDS    10
#define TEST_MSG_1 0x55555555
#define TEST_MSG_2 0x77777777
#define ARINC_573_BUFFER_SIZE 2100
#define INTERRUPT_BUFFER_SIZE 100

static CEI_UINT16 xmitFrameData[ARINC_573_BUFFER_SIZE];
static CEI_UINT16 recvFrameData[ARINC_573_BUFFER_SIZE];

static AR_CHANNEL_INDEX_INFO_TYPE channelConfiguration;        /* board channel configuration structure */
static CDEV_PARM_SSI_TYPE programmableChannelsPresent = 0; /* flag indicating s/w pgm-able chan's avail */
static CEI_CHAR configurationString[40];                                 /* Board name string storage   */
static pCEI_CHAR pConfigurationString;                                   /* Board name string reference */
static pCEI_CHAR altConfigurationString[CEIDEV_NUM_ALT_CONFIGURATIONS] = {" ", "-J", " "};
static pCEI_CHAR discreteChannelState[2] = {"LOW", "HIGH"};  /* displayed detected discrete input state */
static CEI_CHAR filterTable[MAX_ESSM][MAX_SDI][MAX_LABEL];     /* local label filter table content      */
static CDEV_CHAN_TYPE numTxChan;                               /* Number of transmitters installed      */
static CDEV_CHAN_TYPE numRxChan;                               /* Number of receivers installed         */
static CDEV_PARM_SSI_TYPE wrapMode = AR_WRAP_ON;               /* Internal wrap state, on by default    */
static CDEV_PARM_SSI_TYPE disable = AR_ON;                     /* Transmit disable state, on by default */
CEI_UINT32 globalInterruptCounter = 0;                                   /* number of interrupts logged */
CEI_UINT32 globalQueueEntries[INTERRUPT_BUFFER_SIZE];                       /* storage for intQ entries */

/*===========================================================================*
 * ENTRY POINT:                      M A I N
 *===========================================================================*
 *
 * FUNCTION:    Main entry point for the example internal wrap test.
 *
 * DESCRIPTION: This is a simple test program that sends out some data on a
 *              transmit channel and reads it back in on the respective  
 *              receive channel(s).
 *
 *===========================================================================*/
#ifdef __TURBOC__
   #pragma argsused
#endif

#if defined (VXW_PCI_X86) || defined (VXW_PCI_PPC)
CEI_VOID wrap(CEI_VOID)
#elif defined (_LINUX_X86_) || defined (_LINUX_PPC_)
CEI_NATIVE_INT main(CEI_NATIVE_INT argc, pCEI_CHAR argv[])
#else
CEI_VOID main(CEI_NATIVE_INT argc, pCEI_CHAR argv[])
#endif
{
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
#if !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) 
   CEI_NATIVE_INT result;             /* status from scanf operator prompt   */
   CEI_NATIVE_INT argIndex;           /* Index of arguments passed to main() */
#endif                              
   CEI_NATIVE_INT deviceNum;          /* User-entered WinRT/Condor Device ID */
#endif   
   CDEV_PARM_SSI_TYPE fwVersion;      /* board's firmware version            */
   CDEV_PARM_SSI_TYPE configuration;  /* The board's model/configuration     */
   CDEV_PARM_SSI_TYPE irigAvailable;  /* Availability of IRIG on the board   */
   CEI_NATIVE_INT numErrors      = 0; /* Number of errors detected           */
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
   CEI_NATIVE_INT restoreConfig  = 0; /* Restore original x30 PCI BAR2 size  */
#endif   
   CEI_NATIVE_INT advTestSelect  = 0; /* Advanced API test selection         */
   CEI_NATIVE_INT irigDemoSelect = 0; /* IRIG demonstration selection        */
   CEI_NATIVE_INT dioDemoSelect = 0;  /* Discrete I/O demonstration selection*/
   CEI_NATIVE_INT pciIntDemoSelect = 0; /* H/W Interrupt demo selection      */
   CDEV_BOARD_TYPE board = MAX_BOARDS;  /* User-entered board/Device ID      */
   CDEV_API_RET_TYPE status = ARS_NORMAL;  /* API function return status     */

   ar_version (configurationString);
   printf ("\n");
   printf ("     --------------------------------------------------------------\n");
   printf ("           CEI-x30 Basic Internal Wrap Test  Version %s\n", configurationString);
   printf ("     --------------------------------------------------------------\n");
   printf ("       This program performs a basic internal wrap test of each    \n");
   printf ("       paired transmit channel to the respective receive channel.  \n");
   printf ("       The results will be displayed in this console window.       \n");
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
   printf ("       Optional program arguments include:                         \n");
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) 
   printf ("          Any Numeric Value - defines the device ID to reference   \n");
   printf ("       Any other character following the device ID is evaluated as:\n");
   printf ("          A - invokes a demonstration of advanced API routines     \n");
   printf ("          I - invokes a demonstration of the IRIG features         \n");
   printf ("          D - invokes a demonstration of Discrete I/O features     \n");
   printf ("          P - invokes a demonstration of the PCI Interrupt features\n");
   printf ("          other - executes with external transmit & receive enabled\n");
#else
   printf ("         /BD   - provides for an optional device ID to reference   \n");
   printf ("         /ADV  - invokes a demonstration of advanced API routines  \n");
   printf ("         /IRIG - invokes a demonstration of the IRIG features      \n");
   printf ("         /DIO  - invokes a demonstration of Discrete I/O features  \n");
   printf ("         /INT  - invokes a demonstration of PCI Interrupt features \n");
   printf ("         /EXT  - executes with external transmit & receive enabled \n");
#endif   
#endif   
   printf ("     --------------------------------------------------------------\n");

   /*------------------------------------------------------------------------*
    * Process the possible argument values:
    * 1) Include the ADVANCED API test execution (not executed by default)
    * 2) Include the IRIG Features test execution (not executed by default)
    * 3) Include the Discrete I/O test execution (not executed by default)
    * 4) Externally transmit, disable the receiver internal wrap operation
    * 5) Alternate Board Number/Device ID (prompt operator if not specified)
    *------------------------------------------------------------------------*/
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (INTEGRITY_PCI_PPC)
#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) 
   if (argc == 2) {
      sscanf (argv[1], "%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
   }
   else if (argc == 3) {
      char entry;
      sscanf (argv[1], "%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
      sscanf (argv[2], "%c", &entry);
      if (entry == 'A') {
         advTestSelect = 1;
      }
      else if (entry == 'I') {
         irigDemoSelect = 1;
      }
      else if (entry == 'D') {
         dioDemoSelect = 1;
      }
      else if (entry == 'P') {
         pciIntDemoSelect = 1;
      }
      else if (entry == 'R') {
         restoreConfig = 1;
      }
      else {
         wrapMode = AR_WRAP_OFF;
         disable = AR_OFF;     
      }
   }
   else
      board = 0;
#else
   for (argIndex = 1; argIndex < argc; argIndex++) {
      if (strnicmp ("/ADV", argv[argIndex], 4) == 0) {
         advTestSelect = 1;
      }
      else if (strnicmp ("/IRIG", argv[argIndex], 5) == 0) {
         irigDemoSelect = 1;
      }
      else if (strnicmp ("/DIO", argv[argIndex], 4) == 0) {
         dioDemoSelect = 1;
      }
      else if (strnicmp ("/R", argv[argIndex], 2) == 0) {
         restoreConfig = 1;
      }
      else if (strnicmp ("/INT", argv[argIndex], 4) == 0) {
         pciIntDemoSelect = 1;
      }
      else if (strnicmp ("/EXT", argv[argIndex], 4) == 0) {
         wrapMode = AR_WRAP_OFF;
         disable = AR_OFF;     
      }
      else if (strnicmp ("/BD", argv[argIndex], 3) == 0) {
         sscanf (argv[argIndex]+3, "%d", &board);
         if (board >= MAX_BOARDS) {
            printf ("\n\nERROR on Board ID parameter selection /BD%d\n\n", board);
         }
      }      
   }

   /*------------------------------------------------------------------------*
    *  If not defined via argument, ask the user for the board/device number.
    *------------------------------------------------------------------------*/
   if (board >= MAX_BOARDS) {
      printf ("Enter the board number (installation device number): ");
      result = scanf("%d", &deviceNum);
      board = (CDEV_BOARD_TYPE)deviceNum;
      fflush(stdin);
   }
#endif
   /* check for the ability of this device to support the enhanced CEI-x30 firmware;
    * or reprogram/restore to standard firmware BAR2 size.
    */
   check_enhanced_firmware_capability (board, restoreConfig);

#else
   board = 0;    /* default device index for VxWorks and Integrity is zero */
#endif

   /* uncomment to execute the internal wrap test performed during initialization */
/*   ar_bypass_wrap_test ( board, AR_OFF ); 
*/   
   /*------------------------------------------------------------------------*
    *  Initialize the board, exit if not found or it fails initialization.
    *------------------------------------------------------------------------*/
   status = ar_open (board);          /* pass the Device Number of the board */

   if (status != ARS_NORMAL)
      error_exit (status);

   /* acquire the board configuration/type */
   status = ar_get_device_config (board, 0, ARU_CONFIGURATION, &configuration);
   status = ar_get_device_config (board, 0, ARU_FW_VERSION, &fwVersion);
   status = ar_get_device_config (board, 0, ARU_IRIG_AVAILABLE, &irigAvailable);

   /* acquire the channel configuration to determine if it contains s/w programmable channels */
   status = ar_get_device_config (board, 0, ARU_SPECIAL_PROGRAMMABLE_CONFIG, &programmableChannelsPresent);
   
   /* while not necessary for this example's execution, you can demonstrate
    * multi-thread protection at the API level by switching AR_OFF to AR_ON. */
   ar_set_multithread_protect(board, AR_OFF);
   
   /*------------------------------------------------------------------------*
    *  Read the number of channels installed on the board
    *------------------------------------------------------------------------*/
   numTxChan = ar_num_xchans ( board );     /* Total number of transmitters  */
   numRxChan = ar_num_rchans ( board );     /* Total number of receivers     */
   
   /*------------------------------------------------------------------------*
    *  Display the detected CEI-x30 board type and channel configuration
    *------------------------------------------------------------------------*/
   if (configuration & CEIDEV_ALT_CONFIGURATION_717) {  
      numTxChan--;              /* -J module part # doesn't include 717 chan */
      numRxChan--;
   }
   
   pConfigurationString = ar_get_boardname (board, configurationString);
   
   printf ( "This program has detected a");
   if ((configuration & CEIDEV_CONFIGURATION_MASK) == CEIDEV_CONFIG_R830RX) {
      printf ( " n %s", configurationString );
   }
   /* Non-combo s/w programmable channel configurations get special channel config extension part number convention */
   else if ( ((configuration & CEIDEV_CONFIGURATION_MASK) == CEIDEV_CONFIG_RAR_XMC) && programmableChannelsPresent) {
      printf ( "n %s-", configurationString );
      switch ((CEI_UINT32)programmableChannelsPresent) {
         case RAR_XMC_16P16_SPECIAL_CFG:
            printf ( "16P16" );
            break;
         case RAR_XMC_15P15_SPECIAL_CFG:
            printf ( "15P15" );
            break;
         case RAR_XMC_P16_SPECIAL_CFG:
            printf ( "ALT-P16" );
            break;
         default:
            printf ( "unknown" );
            break;
      }
   }
   else {
      printf ( " %s-%02d%02d%s", configurationString, (CEI_NATIVE_INT)numRxChan, (CEI_NATIVE_INT)numTxChan, 
               altConfigurationString[(configuration & CEIDEV_ALT_CONFIGURATION_717) >> CEIDEV_ALT_CONFIG_717_SHIFT]);
   }
   if (irigAvailable) {
      printf ( " with IRIG" );
   }
   printf ( "  F/W v%d.%d\n\n", (CEI_INT32)(fwVersion >> 4), (CEI_INT32)(fwVersion & 0x0F) );

   /*------------------------------------------------------------------------*
    *  Execute API and hardware installation verification test(s)
    *------------------------------------------------------------------------*/

   /* Verify ARINC 429 Data Wrap for all configurations using one-shot msgs  */
   if (configuration != CEIDEV_CONFIG_R830RX) {
      status = test_basic_arinc_429 ( board, &numErrors );
   }
   
   /*  If installed, verify ARINC 573/717 Frame Data Wrap */
   if ( (status == ARS_NORMAL) && (configuration & CEIDEV_ALT_CONFIGURATION_717) ) {  
      status = test_arinc_717 ( board, &numErrors );
   }

   /*------------------------------------------------------------------------*
    * Conditionally execute demonstrations of advanced board functionality
    *------------------------------------------------------------------------*/
   
   /* Next demonstrate label filtering, message scheduling, and snapshot rx */
   if (advTestSelect)
      status = demo_advanced_arinc_429 ( board, &numErrors );

   /* Next demonstrate PCI Interrupt feature usage */
   if (pciIntDemoSelect)
      status = demo_pci_interrupts ( board );

   /* Next demonstrate Discrete I/O utility usage */
   if (dioDemoSelect)
      status = demo_discrete_io_features ( board, &numErrors );

   /* Next demonstrate various IRIG and internal timer API utilities */
   if (irigDemoSelect && irigAvailable)
      status = demo_irig_features ( board );
      
   /*------------------------------------------------------------------------*
    *  Release all resources acquired when this board was first accessed.
    *------------------------------------------------------------------------*/
   status = ar_close (board);

   /*------------------------------------------------------------------------*
    *  Now that we are done, print out final status.
    *------------------------------------------------------------------------*/
   if (status != ARS_NORMAL)
      printf("\nLast Error detected: %s\n", ar_get_error (status));

   if ( numErrors )
      printf("\nTest unsuccessful, %d errors were detected.\n", (CEI_NATIVE_INT)numErrors);
   else
      printf("\nTest PASSED, no errors were detected.\n");

#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC)
   printf("Press any key to exit\n");
   getch();
#endif

#if defined (_LINUX_X86_) || defined (_LINUX_PPC_) 
   return 0;
#endif
}


/*===========================================================================*
 * ENTRY POINT:       T E S T _ B A S I C _ A R I N C _ 4 2 9
 *===========================================================================*
 *
 * FUNCTION:    Perform a simple wrap test on the ARINC 429 channels
 *
 * DESCRIPTION: This is a simple test routine that transmits 10 ARINC 429 
 *              messages with internal wrap enabled.  The transmit and
 *              receive channels are setup for a bus speed 12.5kbps and ODD 
 *              parity.  Generic Condor ARINC 429 API functions are utilized.
 *
 *===========================================================================*/

CDEV_API_RET_TYPE test_basic_arinc_429 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount )
{
   ARINC word;                      /* counter for which word we have sent   */
   ARINC arincMessage;              /* ARINC word that we will send          */
   ARINC receivedMessage;           /* ARINC word that we received           */
   CEI_UINT32 timeTag;              /* Time Tag assigned to received data    */
   CDEV_CHAN_TYPE channel;          /* Current transmit channel being tested */
   CDEV_CHAN_TYPE numMatchedChannels;     /* # of matched Tx/Rx channels     */
   CDEV_API_RET_TYPE status = ARS_NORMAL; /* utility status return value     */
   CEI_UINT32 i;                          /* generic index                   */

   /* Test Matched Rx/Tx channel count */
   numMatchedChannels = (CDEV_CHAN_TYPE)((numTxChan <= numRxChan) ? numTxChan : numRxChan);
   
   /*------------------------------------------------------------------------*
    *  Set the receiver channels to the low (12.5K) bit rate & ODD parity,
    *  also enable internal wrap.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < numRxChan) && (status == ARS_NORMAL); channel++) {
      /* enable internal wrap  */
      status = ar_set_device_config ( board, channel, ARU_RECV_MODE, wrapMode );

      /* receive @ 12.5kbs */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, ARU_SPEED_LOW );
      
      /* parity check enabled for receive channels */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_PARITY, AR_ON );

      /* use individual buffered FIFO */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_MERGED_MODE, AR_OFF );

      /* enable reception */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_ON );
   }
   
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_set_device_config() Rx channel %d: %s\n", (CEI_NATIVE_INT)channel, ar_get_error(status) );
      (*errorCount)++;
   }

   /*------------------------------------------------------------------------*
    *  If software programmable channels are present, assign them to be
    *  transmitters for this example.  You can also index the programmable
    *  by transmit channel index, since programmable channel boards always
    *  allocate all transmitters as programmable channels.
    *------------------------------------------------------------------------*/
   if (programmableChannelsPresent) {
      status = ar_get_channel_index_info (board, &channelConfiguration);
      for (i = 0; i < channelConfiguration.num_channels; i++) {
         if (channelConfiguration.chan_type[i] == AR_CHAN_TYPE_429_RX_OR_TX) {
            status = ar_set_device_config ( board, (CEI_INT16)channelConfiguration.chan_tx_index[i], ARU_CONFIG_PROGRAMMABLE_CHAN, ARU_TRANSMITTER );
         }
      }
   }

   /*------------------------------------------------------------------------*
    *  Set the transmitter channels to the low (12.5K) bit rate & ODD parity.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < numTxChan) && (status == ARS_NORMAL); channel++) {
      /* transmit @ 12.5kbs */
      status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, ARU_SPEED_LOW );

      /* odd parity for both transmit channels */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_TX_PARITY, ARU_PARITY_ODD ); 

      /* disable external transmission */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_TX_DISABLE, disable );

      /* for now disable transmit FIFO */
      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_TX_FIFO_ENABLE, AR_OFF );
   }
   
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_set_device_config() Tx channel %d: %s\n", (CEI_NATIVE_INT)channel, ar_get_error(status) );
      (*errorCount)++;
   }


   /*------------------------------------------------------------------------*
    *  Activate message processing.
    *------------------------------------------------------------------------*/
   if (status == ARS_NORMAL) {
      status = ar_go (board);

      if (status != ARS_NORMAL) {
         printf ("Error status from ar_go(): %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
   }

   /*------------------------------------------------------------------------*
    *  Stuff incrementing data in the transmit buffer then enable the FIFO.
    *
    *  NOTE:   We don't want to overflow the transmit buffer.  For large
    *          quantities of data, the return status from transmit functions 
    *          should be checked for overflow.  Failure indicates the FIFO is  
    *          full at this time and the application should retry later.
    *------------------------------------------------------------------------*/
   timeTag = 0;
   for (channel = 0; (channel < (numMatchedChannels)) && (status == ARS_NORMAL); channel++) {

      printf ( "429 Transmit Channel %d is wrapped to 429 Receive Channel %d\n",(CEI_NATIVE_INT)channel, (CEI_NATIVE_INT)channel );

      /* transmit N_WORDS messages */
      for ( word = 0; (word < N_WORDS) && (status == ARS_NORMAL); word++ ) {
         arincMessage = (word * (ARINC)0x01111111);
         status = ar_putword (board, channel, (CDEV_PARM_SI_TYPE)arincMessage);
         if (status != ARS_NORMAL) {
            printf ( "Error status from ar_put_429_message() channel %d: %s\n", (CEI_NATIVE_INT)channel, ar_get_error(status) );
            (*errorCount)++;
         }
      }
      
      status = ar_set_device_config ( board,       /* enable transmission */
                                      channel, 
                                      ARU_TX_FIFO_ENABLE, 
                                      AR_ON );

      /*---------------------------------------------------------------------*
       *  Read back the data we just transmitted.
       *---------------------------------------------------------------------*/
      status = ARS_GOTDATA;
      for ( word = 0; (word < N_WORDS) && (status == ARS_GOTDATA); word++ ) {
         arincMessage = (word * (ARINC)0x01111111);  /* expected data pattern */

         status = ar_getnextt ( board, channel, (ARINC *)&receivedMessage, (pCEI_UINT32)&timeTag );

         if (status == ARS_GOTDATA) {
            if ( arincMessage == receivedMessage ) {
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
               printf ("Xmit: %08X Recvd: %08X on Channel %d :: Timetag: %08X Success!\n", 
                       arincMessage, 
                       receivedMessage,
                       (CEI_NATIVE_INT)channel,
                       timeTag );
#endif
            }
            else {
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
               printf ("Xmit: %08X Recvd: %08X  Channel %d :: PATTERN MISMATCH!\n",
                       arincMessage, 
                       receivedMessage, 
                       (CEI_NATIVE_INT)channel );
#endif
               (*errorCount)++;
            }
         }
         else {
            printf ("Channel %d Timeout on Data Reception\n", (CEI_NATIVE_INT)channel );
            (*errorCount)++;
         }
      }

      if ( (channel < numMatchedChannels-1) && (status == ARS_GOTDATA) ) {
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
         printf("\nEnter any key to test the next transmit channel\n");
         getch();
#endif
      }

      if ((status == ARS_GOTDATA) || (wrapMode == AR_WRAP_OFF))
         status = ARS_NORMAL;
   }
   
   if (status == ARS_NORMAL)
      status = ar_stop (board);
   else {
      (*errorCount)++;
      ar_stop (board);
   }

   if (*errorCount)
      printf ( "\nThe ARINC 429 Internal Wrap Test encountered %d errors.\n", (CEI_NATIVE_INT)*errorCount);
   else
      printf ( "\nThe ARINC 429 Internal Wrap Test executed successfully.\n");

   return status;
}

/*============================================================================*
 * ENTRY POINT:     D E M O _ A D V A N C E D _ A R I N C _ 4 2 9
 *============================================================================*
 *
 * FUNCTION:    Demonstrate the advanced features supporting ARINC 429
 *
 * DESCRIPTION: This is a demonstration of the following CEI-x30 features:
 *              
 *                 Transmit Message Scheduling
 *                 Enhanced Label/SDI/ESSM Data Filtering
 *                 Snapshot Message Data Acquisition
 *                 Enhanced Time-tag Reset and Conversion
 *                 IRIG Time-tag Selection (if installed on hardware)
 *
 *============================================================================*/
CDEV_API_RET_TYPE demo_advanced_arinc_429 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount )
{
   CEI_UINT32 msgIndex;            /* define message index                    */
   CEI_UINT32 receivedMessage;     /* ARINC word that we received             */
   CEI_UINT32 snapshotMessage = 0; /* ARINC word that we received             */
   CEI_UINT32 timeTag;             /* Time Tag assigned to received data      */
   CDEV_CHAN_TYPE channel;         /* Current transmit channel being tested   */
   CDEV_CHAN_TYPE numMatchedChannels;    /* Number of matched Tx/Rx channels  */
   CDEV_API_RET_TYPE status = ARS_NORMAL;/* Utility status return value       */
   CDEV_API_RET_TYPE msgAssignment;  /* Msg Index returned from ar_define_msg */
   CEI_CHAR  seq_num;            /* Legacy ar_get_latest parameter (not used) */
   CEI_INT32 labelIndex;         /* Temporary label filter table index        */
   CEI_INT32 sdiIndex;           /* Temporary label filter table index        */
   CEI_INT32 ssmIndex;           /* Temporary label filter table index        */
   CDEV_PARM_SSI_TYPE irigStatus = 0; /* Current validity of the IRIG signal  */
   CDEV_PARM_SSI_TYPE irigAvail; /* availability of IRIG on the device used   */
   AR_TIMETAG_TYPE timeTagStruct;     /* Time-tag structure for ar_getnext_xt */
   CEI_CHAR displayTime[AR_TD_STRING_SIZE]; /* Msg Time-tag convert to string */
   pCEI_CHAR filterTblPtr = &filterTable[0][0][0];/* Fltr tbl for ar_putfilter*/
   AR_SCHEDULED_MSG_ENTRY_TYPE messageEntry[N_WORDS]; /* Message table for 
                                                         ar_define_msg_block */
   printf ( "\n\nNow Executing the Enhanced CEI-x30 API Demonstration.\n\n");
   
   /*------------------------------------------------------------------------*
    * Set the 64-bit 1 microsecond device timer to an initial value of 50 
    * seconds.  Then retrieve the current time, convert it to a string,
    * and display it in the console window.
    *------------------------------------------------------------------------*/
   timeTagStruct.timeTag = 50000000L;
   timeTagStruct.timeTagFormat = AR_TIMETAG_INT_USEC_64BIT;
   status = ar_set_time (board, &timeTagStruct);

   /* If IRIG is available, demonstrate how to reference it as well */
   status = ar_get_device_config (board, 0, ARU_IRIG_AVAILABLE, &irigAvail);

   if ( (status == ARS_NORMAL) && irigAvail ) {

      /* enable IRIG receiver internal wrap from IRIG generator (board-wide, not channel selectable) */
      status = ar_set_device_config (board, 0, ARU_IRIG_WRAP_ENABLE, (CDEV_PARM_SSI_TYPE)disable);
      
      /*------------------------------------------------------------------------*
       * if any other equipment's IRIG time deviates from the CEI-x30,
       * offset the x30 IRIG time value by (set in milliseconds)
       * note again, IRIG status/operation is board-wide, not channel selectable
       *------------------------------------------------------------------------*/
      status = ar_set_device_config (board, 0, ARU_IRIG_SET_BIAS, 0);
      
      /* select the IRIG receiver as the system time source (board-wide, not channel selectable) */
      status = ar_set_device_config (board, 0, ARU_RX_TIMETAG_MODE, AR_TIMETAG_EXT_IRIG_64BIT);     

      /* wait for successive received valid IRIG time values */
      while (!irigStatus) {
         status = ar_get_device_config (board, 0, ARU_IRIG_CALIBRATED, &irigStatus);
      }
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_64BIT, &timeTagStruct);
   }
   else {
      status = ar_get_time (board, AR_TIMETAG_INT_USEC_64BIT, &timeTagStruct);
   }
   
   ar_convert_time_to_string (board, AR_TD_DATE, &timeTagStruct, displayTime);
   printf ("Starting the test at time: %s\n\n", displayTime );

   /* Test Matched Rx/Tx channel count */
   numMatchedChannels = (CDEV_CHAN_TYPE)((numTxChan <= numRxChan) ? numTxChan : numRxChan);
   
   /*------------------------------------------------------------------------*
    *  Set the transmit channels to the high (100K) bit rate & disable parity.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < numTxChan) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, ARU_SPEED_HIGH );

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_TX_PARITY, ARU_PARITY_NONE ); 

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_TX_FIFO_ENABLE, AR_ON );
   }
   
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_set_device_config() Tx channel %d: %s\n", (CEI_NATIVE_INT)channel, ar_get_error(status) );
      (*errorCount)++;
   }

   /*------------------------------------------------------------------------*
    *  Set the receive channels to the high (100K) bit rate & disable parity,
    *  also enable internal wrap.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < numRxChan) && (status == ARS_NORMAL); channel++) {

      status = ar_set_device_config ( board, channel, ARU_RECV_MODE, wrapMode );

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, ARU_SPEED_HIGH );

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_PARITY, AR_OFF ); 

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_MERGED_MODE, AR_OFF );

      if (status == ARS_NORMAL)
         status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_ON );
   }
   
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_set_device_config() Rx channel %d: %s\n", (CEI_NATIVE_INT)channel, ar_get_error(status) );
      (*errorCount)++;
   }
   else {
      /*------------------------------------------------------------------------*
       * Setup the Snapshot Storage Mode to save all message label data in the
       * snapshot buffer based only on the label value (not label+SDI).
       *------------------------------------------------------------------------*/
      status = ar_set_device_config ( board, 
                                      0,      /* channel number is "don't care" */
                                      ARU_ACCESS_SNAPSHOT_BUFFER, 
                                      ARU_LABEL_ONLY );
   }
   
   /*------------------------------------------------------------------------------*
    * Create some periodic transmit messages using AR_DEFINE_MSG_BLOCK 
    * Define N_WORDS of messages across varying Tx channels at a 100ms (10Hz) rate.
    * Define and increment the label value for each message using the index value.
    *------------------------------------------------------------------------------*/
   for (msgIndex = 0; msgIndex < N_WORDS; msgIndex++) {
      messageEntry[msgIndex].board   = (CEI_UINT32)board;          /* vary Tx Chan */
      messageEntry[msgIndex].channel = (CEI_UINT32)(msgIndex % numMatchedChannels); 
      messageEntry[msgIndex].rate    = 100;
      messageEntry[msgIndex].start   = 0;
      messageEntry[msgIndex].txCount = CEIDEV_MSG_SCHEDULER_TX_ALWAYS;
      messageEntry[msgIndex].data    = 0x12345601 + msgIndex;
   }

   status = ar_define_msg_block (N_WORDS, &messageEntry[0]);
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_define_msg_block() %s\n", ar_get_error(status) );
      (*errorCount)++;
   }

   /*------------------------------------------------------------------------*
    * Now create two more 50ms periodic transmit messages on transmit  
    * channel 0 using AR_DEFINE_MSG.  Fix the pattern for Label Filtering
    * and Snapshot Buffer usage demonstration.
    *------------------------------------------------------------------------*/
   channel = 0;

   msgAssignment = ar_define_msg (board, channel, 50, 0, TEST_MSG_1);
   if (msgAssignment != N_WORDS) {
      printf ("Message index from ar_define_msg() not what was expected: %d\n",
              (CEI_NATIVE_INT)msgAssignment );
      (*errorCount)++;
   }
   
   msgAssignment = ar_define_msg (board, channel, 50, 0, TEST_MSG_2);

   /*------------------------------------------------------------------------*
    * Assign a best-fit start offset value to the defined scheduled message
    * scenario
    *------------------------------------------------------------------------*/
   status = (CEI_INT16)ar_assign_scheduler_start_offsets(board);
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_assign_scheduler_start_offsets() %s\n", ar_get_error(status) );
      (*errorCount)++;
   }

   /*------------------------------------------------------------------------*
    * Setup the receive channel 0 label filter table to filter all 
    * label/SDI/ESSM combinations from reception.
    * This could also be accomplished via ar_enh_label_filter() as follows:
    *   ar_enh_label_filter (board, 
    *                        channel, 
    *                        ARU_ALL_LABELS, 
    *                        ARU_ALL_SDI, 
    *                        ARU_ALL_ESSM, 
    *                        FILTER_SEQUENTIAL + FILTER_SNAPSHOT);
    *------------------------------------------------------------------------*/
   for (ssmIndex = 0; ssmIndex < MAX_ESSM; ssmIndex++) {
      for (sdiIndex = 0; sdiIndex < MAX_SDI; sdiIndex++) {
         for (labelIndex = 0; labelIndex < MAX_LABEL; labelIndex++) {
            filterTable[ssmIndex][sdiIndex][labelIndex] = FILTER_SEQUENTIAL + FILTER_SNAPSHOT;
         }
      }
   }
   status = (CEI_INT16)ar_putfilter (board, channel, filterTblPtr);
   if (status != ARS_NORMAL) {
      printf ("Error status from ar_putfilter(): %s\n", ar_get_error(status) );
   }

   /*------------------------------------------------------------------------*
    * disable label filtering for the two messages defined via ar_define_msg 
    *
    *    label = value of bit field b07-b00
    *    sdi   = value of bit field b09-b08
    *    essm  = value of bit field b30-b28
    *
    * action value of 0 enables snapshot and sequential receive buffer storage
    *
    *------------------------------------------------------------------------*/
   status = (CEI_INT16)ar_enh_label_filter (board, 
                                            channel, 
                                            (CEI_UINT16)( TEST_MSG_1 & 0xff), 
                                            (CEI_UINT16)((TEST_MSG_1 & 0x300) >> 8), 
                                            (CEI_UINT16)((TEST_MSG_1 & 0x70000000) >> 28), 
                                            0);
   if (status == ARS_NORMAL)
      status = (CEI_INT16)ar_enh_label_filter (board, 
                                               channel, 
                                               (CEI_UINT16)( TEST_MSG_2 & 0xff), 
                                               (CEI_UINT16)((TEST_MSG_2 & 0x300) >> 8), 
                                               (CEI_UINT16)((TEST_MSG_2 & 0x70000000) >> 28), 
                                               0);

   if (status != ARS_NORMAL) {
      printf ("Error status from ar_enh_label_filter(): %s\n", ar_get_error(status) );
      (*errorCount)++;
   }

   /*------------------------------------------------------------------------*
    *  Start the board running.
    *------------------------------------------------------------------------*/
   if (status == ARS_NORMAL) {
      status = ar_go (board);

      if (status != ARS_NORMAL) {
         printf ("Error status from ar_go(): %s\n", ar_get_error(status) );
         (*errorCount)++;
      }

      /* insert some messages that should be filtered... */
      status = ar_putword (board, channel, (TEST_MSG_1 & 0xfffff0ff));
      status = ar_putword (board, channel, (TEST_MSG_2 & 0xfffff0ff));
      status = ar_putword (board, channel, (TEST_MSG_1 & 0xfffffffe));
      status = ar_putword (board, channel, (TEST_MSG_2 & 0xfffffffe));
      status = ar_putword (board, channel, (TEST_MSG_1 & 0x8fffffff));
      status = ar_putword (board, channel, (TEST_MSG_2 & 0x8fffffff));

      /*---------------------------------------------------------------------*
       * Check receiver 0 for the unfiltered labels, wait for first message.
       *---------------------------------------------------------------------*/
      /* demo ar_get_429_message() to retrieve the first message (using a delay) */
      status = ar_get_429_message (board, channel, AR_ON, &receivedMessage, &timeTag);
      
      if (status == ARS_GOTDATA) {
         printf ("Received Data for First Msg: 0x%08X at time: %d microseconds\n", (CEI_UINT32)receivedMessage, (CEI_NATIVE_INT)timeTag );
      }
      else {
         printf ("Failed to Receive First Msg on Channel 0: %s\n", ar_get_error(status) );
         (*errorCount)++;
      }

      /*---------------------------------------------------------------------*
       * demo ar_get_latest() to retrieve the first pattern message snapshot
       * entry, passing the label value on the lower 8 bits of the fixed 
       * test message.
       *---------------------------------------------------------------------*/
      ar_get_latest (board, channel, (CDEV_PARM_USI_TYPE)(TEST_MSG_1 & 0xff), &snapshotMessage, &seq_num);
      
      if (snapshotMessage == TEST_MSG_1) {
         printf ("Snapshot Data for First Msg: 0x%08X  Seq: %d\n", (CEI_UINT32)snapshotMessage, (CEI_UINT32)(seq_num & 0xff) );
         ar_sleep (50);
         ar_get_latest (board, channel, (CDEV_PARM_USI_TYPE)(TEST_MSG_1 & 0xff), &snapshotMessage, &seq_num);
         printf ("Snapshot Data for First Msg: 0x%08X  Seq: %d\n", (CEI_UINT32)snapshotMessage, (CEI_UINT32)(seq_num & 0xff) );
      }
      else {
         printf ("Bad Snapshot Entry For First Msg on Channel 0: 0x%08X\n", (CEI_UINT32)snapshotMessage );
         (*errorCount)++;
      }

      /*---------------------------------------------------------------------*
       * Check receiver 0 for the next message
       *---------------------------------------------------------------------*/
      /* demo ar_getnext_xt() to retrieve the second message & convert time */
      status = ar_getnext_xt (board, channel, &receivedMessage, &timeTagStruct);
      if (status == ARS_GOTDATA) {
         ar_convert_time_to_string (board, AR_TD_DATE, &timeTagStruct, displayTime);
         printf ("\nReceived Data for Second Msg: 0x%08X at time: %s seconds\n", (CEI_UINT32)receivedMessage, displayTime );
      }
      else {
         printf ("\nFailed to Receive Second Msg on Channel 0: %s\n", ar_get_error(status) );
         (*errorCount)++;
      }

      /* demo ar_get_snap_data() to retrieve the second message snapshot entry */
      status = (CDEV_API_RET_TYPE)ar_get_snap_data (board, channel, (CDEV_PARM_USI_TYPE)(TEST_MSG_2 & 0xff), 0, &snapshotMessage);
    
      if ((status == ARS_NORMAL) && (snapshotMessage == TEST_MSG_2)) {
         printf ("Snapshot Data for Second Msg: 0x%08X\n", (CEI_UINT32)snapshotMessage );
      }
      else {
         printf ("Encountered either a Bad Status from ar_get_snap_data or \n     An invalid Snapshot Entry For 2nd Msg on Ch 0: 0x%08X\n", snapshotMessage );
         (*errorCount)++;
      }
      
      /*------------------------------------------------------------------------*
       * Now modify the message data values for those periodic transmit messages 
       * using AR_MODIFY_MSG_BLOCK 
       *------------------------------------------------------------------------*/
      for (msgIndex = 0; msgIndex < N_WORDS; msgIndex++) {
         messageEntry[msgIndex].data    = 0x65432101 + msgIndex;
      }
      status = ar_modify_msg_block (N_WORDS, &messageEntry[0]);
      if (status != ARS_NORMAL) {
         printf ("Error status from ar_modify_msg_block() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
   }

   if (status == ARS_NORMAL)
      status = ar_stop (board);
   else {
      (*errorCount)++;
      ar_stop (board);
   }

#if 0
   /* Demonstrate Access to the RAR-PCIE Hardware Monitor Features */
   configuration = (CEI_INT32)ar_get_boardtype (board);
   if ((configuration == CEI430A) || (configuration == RARPCIE)) 
   {
      if (configuration == CEI430A) {
         CDEV_PARM_SSI_TYPE fpgaTemp;
         status = ar_get_device_config (board, 0, ARU_HW_FPGA_TEMPERATURE, &fpgaTemp);
         printf ("\nCEI-430A FPGA Temperature is %d deg C\n", fpgaTemp);
      }
      else {
         CDEV_PARM_SSI_TYPE fpgaTemp, oneVoltSupply, twoPtFiveVoltSupply;
         status = ar_get_device_config (board, 0, ARU_HW_FPGA_TEMPERATURE, &fpgaTemp);
         status = ar_get_device_config (board, 0, ARU_HW_1PT0V_PWR_SUPPLY, &oneVoltSupply);
         status = ar_get_device_config (board, 0, ARU_HW_2PT5V_PWR_SUPPLY, &twoPtFiveVoltSupply);
         printf ("\nRAR-PCIE FPGA Temperature is %2.3f deg C\n", (CEI_FLOAT)(fpgaTemp) / (CEI_FLOAT)(1000));
         printf (  "RAR-PCIE FPGA 1.0V Supply is %1.3f Volts\n", (CEI_FLOAT)(oneVoltSupply) / (CEI_FLOAT)(1000));
         printf (  "RAR-PCIE FPGA 2.5V Supply is %1.3f Volts\n", (CEI_FLOAT)(twoPtFiveVoltSupply) / (CEI_FLOAT)(1000));
      }
   }
#endif

   return status;
}

/*============================================================================*
 * ENTRY POINT:    D E M O _ D I S C R E T E _ I O _ F E A T U R E S
 *============================================================================*
 *
 * FUNCTION:    Demonstrate the API Utilities for Discrete I/O
 *
 * DESCRIPTION: This is a demonstration on the use of the Discrete I/O
 *              Channels and the respective API routines.
 *
 *============================================================================*/
CDEV_API_RET_TYPE demo_discrete_io_features ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount )
{
   CDEV_API_RET_TYPE status = ARS_NORMAL;     /* Utility status return value  */
   CDEV_PARM_SSI_TYPE channelCount;
   CDEV_CHAN_TYPE discreteChannel;
   CDEV_PARM_SSI_TYPE discreteState;

   status = ar_get_device_config (board, 0, ARU_CHAN_COUNT_DISC, &channelCount);
   if (channelCount > 0) {
      printf ( "\n\nNow Executing the CEI-x30 Discrete I/O Demonstration.\n\n");
      printf ( "Detected %d Discrete Channels.\n", (CEI_NATIVE_INT)channelCount);

      /* set each Discrete Output to High */
      for (discreteChannel = 0; discreteChannel < (CDEV_CHAN_TYPE)channelCount; discreteChannel++) {
         status = ar_set_device_config (board, discreteChannel, ARU_DISCRETE_OUT, AR_HI);
         if (status != ARS_NORMAL) {
            printf ("\nError setting discrete %d: %s ", (CEI_NATIVE_INT)discreteChannel, ar_get_error(status)); 
            (*errorCount)++;
         }
         else
            printf ( "\nDiscrete Output Channel %d set HIGH", (CEI_NATIVE_INT)discreteChannel);
      }
      /* delay for the Discrete Output setting to take effect */
      ar_wait ( (float)(0.01) );

      /* If the Discrete Output were tied to the Discrete Input, a state change would be detected */
      for (discreteChannel = 0; discreteChannel < (CDEV_CHAN_TYPE)channelCount; discreteChannel++) {
         status = ar_get_device_config (board, discreteChannel, ARU_DISCRETE_IN, &discreteState);
         if (status != ARS_NORMAL) {
            printf ("\nError reading discrete %d: %s ", (CEI_NATIVE_INT)discreteChannel, ar_get_error(status)); 
            (*errorCount)++;
         }
         else
            printf ( "\nDiscrete Input Channel %d set to %s", (CEI_NATIVE_INT)discreteChannel, discreteChannelState[discreteState]);
      }
   }
   return status;
}

/*============================================================================*
 * ENTRY POINT:            D E M O _ I R I G _ F E A T U R E S
 *============================================================================*
 *
 * FUNCTION:    Demonstrate the IRIG features 
 *
 * DESCRIPTION: This is a demonstration of the following CEI-x30 IRIG features:
 *              
 *                 IRIG DAC Threshold Adjustment
 *                 IRIG Bias (Offset) Time Assignment
 *                 IRIG Validity Determination
 *                 IRIG Time Conversion and Display
 *
 *              It requires an external IRIG signal, either from an external
 *              IRIG generator or from the CEI-x30 IRIG generator connected
 *              as described in teh CEI-x30-SW User's Manual.
 *
 *============================================================================*/
CDEV_API_RET_TYPE demo_irig_features ( CDEV_BOARD_TYPE board )
{
   CDEV_API_RET_TYPE status;            /* utility status return value        */
   AR_TIMETAG_TYPE timeTagStruct;       /* Time-tag structure for ar_get_time */
   AR_TIMETAG_TYPE prevTimeTagStruct;   /* Time-tag structure for ar_get_time */
   AR_TIMETAG_TYPE dispTimeTagStruct;   /* Time-tag structure for ar_get_time */
   CEI_CHAR displayTime[AR_TD_STRING_SIZE]; /* Msg TimeTag converted to string*/
   CDEV_PARM_SSI_TYPE configuration;        /* The device model/configuration */
   CDEV_PARM_SSI_TYPE irigStatus = 0;      /* IRIG operability status from get-config */
   CEI_NATIVE_INT lapCounter = 0;

   printf ( "\n\nNow Executing the CEI-x30 IRIG Demonstration.\n\n");

   status = ar_get_device_config (board, 0, ARU_CONFIGURATION, &configuration);
   if (configuration == CEIDEV_CONFIG_R830RX)
      status = ar_set_device_config (board, 0, ARU_IRIG_OUTPUT_ENABLE, AR_ON);

   /*------------------------------------------------------------------------*
    * quick adjust the threshold (mainly for AM signal usage)
    *------------------------------------------------------------------------*/
   status = ar_set_device_config (board, 0, ARU_IRIG_QUICK_ADJUSTMENT, 0);
   if (status == ARS_FAILURE) {
      printf ("\nThe IRIG DAC Adjustment failed to detect an external IRIG signal.\n");
      /* enable IRIG receiver internal wrap from IRIG generator */
      status = ar_set_device_config (board, 0, ARU_IRIG_WRAP_ENABLE, AR_ON);
   }
   else
      printf ("\nThe IRIG DAC Adjustment assigned a threshold of %d millivolts.\n", (CEI_NATIVE_INT)status);

   /*------------------------------------------------------------------------*
    * offset the IRIG time value by 400 milliseconds
    *------------------------------------------------------------------------*/
   status = ar_set_device_config (board, 0, ARU_IRIG_SET_BIAS, 400);

   /*------------------------------------------------------------------------*
    * wait until the IRIG signal is consistently received every 1 second
    *------------------------------------------------------------------------*/
   printf ("\nWaiting for valid IRIG signal.\n");
   while (!irigStatus) {
      status = ar_get_device_config (board, 0, ARU_IRIG_CALIBRATED, &irigStatus);
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_64BIT, &timeTagStruct);
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
      if (kbhit())
         break;
      else
#endif
      ar_wait ((float)0.5);
   }
   
   /*------------------------------------------------------------------------*
    * read IRIG time, convert it for display in a date/time format
    *------------------------------------------------------------------------*/
   status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_64BIT, &timeTagStruct);
   if (status == ARS_NORMAL) {
      ar_convert_time_to_string (board, AR_TD_IRIG, &timeTagStruct, displayTime);
      printf ("\nIRIG Time  %s", displayTime);
   }
   else {
      printf ("\nError status reading IRIG Time: %s\n", ar_get_error(status) );
   }

   /*------------------------------------------------------------------------*
    * track any deviations between consecutive IRIG time values or between
    * the time IRIG time was received and board time
    *------------------------------------------------------------------------*/
   printf ("\n\nPress any key on the keyboard to terminate this demonstration.\n");
   printf ("\n Diff between   Diff between Reference      IRIG  Time");
   printf ("\n IRIG Samples    Time and Sample Time");
   printf ("\n(microseconds)      (microseconds)          (day/time)\n\n");
   status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &prevTimeTagStruct);
   while (status == ARS_NORMAL) {
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_REF_64BIT, &timeTagStruct);
      status = ar_get_time (board, AR_TIMETAG_EXT_IRIG_64BIT, &dispTimeTagStruct);

      /*---------------------------------------------------------------------*
       * If a fresh IRIG time was received, display the difference between 
       * consecutive "IRIG received" time and compare "IRIG received" time 
       * against board reference time.
       *---------------------------------------------------------------------*/
      if (timeTagStruct.timeTag != prevTimeTagStruct.timeTag) {
         ar_convert_time_to_string (board, AR_TD_DATE, &dispTimeTagStruct, displayTime);
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
         printf("\r   %08d            %08d        %s", 
#else
         printf("\n   %08d            %08d        %s", 
#endif
                (CEI_NATIVE_INT)(timeTagStruct.timeTag - prevTimeTagStruct.timeTag),
                (CEI_NATIVE_INT)(timeTagStruct.referenceTimeTag - timeTagStruct.timeTag),
                displayTime);

         prevTimeTagStruct.timeTag = timeTagStruct.timeTag;
         lapCounter++;
      }

#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
      if (kbhit()) {
         printf ("\n");
         getch();
         break;
      }
#else
      if (lapCounter > 10)
         break;
#endif
   }

   return status;
}

/*===========================================================================*
 * ENTRY POINT:         T E S T _ 7 1 7 _ C H A N N E L S
 *===========================================================================*
 *
 * FUNCTION:    Perform an ARINC 573/717 Wrap Test
 * 
 * DESCRIPTION: This is a simple test program that transmits a 573/717 BPRZ  
 *              frame at a rate of 768bps.  The frame consists of 256 words  
 *              with the following 64-word subframe data pattern:
 * 
 *                                    sync  data word 1 ...  data word 63
 *                                   0x123    0x0001    ...    0x003F
 *                                   0x224    0x0041    ...    0x007F
 *                                   0x325    0x0081    ...    0x00BF
 *                                   0x426    0x00C1    ...    0x00FF
 * 
 *              NOTE: This example's transmit and receive data arrays
 *                    xmitFrameData[] and recvFrameData[], are defined to 
 *                    support a total frame size of 2048 words. This limits 
 *                    the subframe size to a maximum of 512 words.  If you 
 *                    desire to use a larger subframe size you must increase 
 *                    the value of ARINC_573_BUFFER_SIZE respectfully.
 * 
 *===========================================================================*/
CDEV_API_RET_TYPE test_arinc_717 ( CDEV_BOARD_TYPE board, CEI_NATIVE_INT * errorCount )
{
   CDEV_PARM_SI_TYPE syncWord[4] = {0x0123, 0x0224, 0x0325, 0x0426};
   CDEV_API_RET_TYPE status = ARS_NORMAL;
   
   CEI_UINT32 subFrameIndex, dataWordIndex; 
   CEI_UINT32 numberWordsToTransmit, numberWordsToReceive;
   CEI_UINT32 subFrameSize, frameSize;
   CEI_INT32 attemptsToWrapFrame = 0;

#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
      printf("\nPress any key to test the ARINC 717 channel\n");
      getch();
#endif

   printf ( "\nThe ARINC 717 Transmit Channel is wrapped to the ARINC 717 Receive Channel\n" );
   printf ( "This test may take several seconds...\n" );

RETRY573:
   /*---------------------------------------------------------------------*  
    * use Bi-polar Return-to-Zero transmission and reception encoding
    *---------------------------------------------------------------------*/
   if (status == ARS_NORMAL)
      status = ar_set_573_config ( board, ARU_TX_FIFO_ENABLE, AR_OFF );
   else {
      printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
      (*errorCount)++;
   }
      
   if (status == ARS_NORMAL)
      status = ar_set_573_config ( board, ARU_TX_DISABLE, disable );
   else {
      printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
      (*errorCount)++;
   }
  
   if (status == ARS_NORMAL) {
      status = ar_set_573_config ( board, ARU_573_RX_BPRZ_SELECT, AR_ON );
      
      if (status != ARS_NORMAL) {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
   }
   else {
      printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
      (*errorCount)++;
   }

   /*---------------------------------------------------------------------*  
    * first flush any existing data from the ARINC 573/717 receiver FIFO
    *---------------------------------------------------------------------*/
   if (status == ARS_NORMAL) {
      status = ar_set_573_config ( board, ARU_573_RX_AUTO_DETECT, AR_OFF );

      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_573_SYNC_WORD1, syncWord[0] );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }

      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_573_SYNC_WORD2, syncWord[1] );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_573_SYNC_WORD3, syncWord[2] );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_573_SYNC_WORD4, syncWord[3] );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_RX_MERGED_MODE, AR_ON );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      /* Define the frame to be 768bps to use with a subframe size of 64 words */
      if (status == ARS_NORMAL) {
         status = ar_set_573_config ( board, ARU_RX_BITRATE, ARU_573_RATE_SIZE_768_64 );
      }
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_RX_FIFO_ENABLE, AR_ON );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL) {
         status = ar_go (board);
      
         if (status == ARS_NORMAL) {
            numberWordsToReceive = 512;
            status = ARS_GOTDATA;
            while ( (status == ARS_GOTDATA) && (numberWordsToReceive > 16) ) {
               status = ar_get_573_frame ( board, &numberWordsToReceive, recvFrameData );
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
               /* terminate the flush loop if operator requests it */
               if ((kbhit()) && (getch()==EXIT_KEY)){
                  return ARS_NORMAL;      
               }
#endif
            }
            status = ar_stop (board);

            if (status == ARS_NORMAL) {
               status = ar_set_573_config ( board, ARU_RX_FIFO_ENABLE, AR_OFF );
               if (status != ARS_NORMAL)
                  printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
            }
            else {
               printf ( "Error status from ar_stop() %s\n", ar_get_error(status) );
               (*errorCount)++;
            }
         }
         else {
            printf ( "Error status from ar_go() %s\n", ar_get_error(status) );
            (*errorCount)++;
         }
      }
   }

   /*---------------------------------------------------------------------*
    * setup the receiver and transmitter configuration registers
    *---------------------------------------------------------------------*/
   if (status == ARS_NORMAL) {
      status = ar_set_573_config ( board, ARU_RECV_MODE, wrapMode );

      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_573_RX_AUTO_DETECT, AR_ON );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }

      if (status == ARS_NORMAL) {
         status = ar_set_573_config ( board, ARU_TX_BITRATE, ARU_573_RATE_SIZE_768_64 );
         if (status != ARS_NORMAL) {
            printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
            (*errorCount)++;
         }
      }
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
   }
   
   subFrameSize = 64;
   frameSize = 4 * subFrameSize;
   
   /* assign the receiver sync words for auto-detect logic */
   if (status == ARS_NORMAL) {
      /* enable the receiver and transmitter FIFOs */
      if (status == ARS_NORMAL)
         status = ar_set_573_config ( board, ARU_TX_FIFO_ENABLE, AR_ON );
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
      if (status == ARS_NORMAL) {
         status = ar_set_573_config ( board, ARU_RX_FIFO_ENABLE, AR_ON );
         if (status != ARS_NORMAL) {
            printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
            (*errorCount)++;
         }
      }
      else {
         printf ("Error status from ar_set_573_config() %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
      
   }
      
   if (status == ARS_NORMAL) {
      status = ar_go (board);
      if (status != ARS_NORMAL)
         printf ( "Error status from ar_go() %s\n", ar_get_error(status) );
   }
   
   /*---------------------------------------------------------------------*  
    * define the frame data array with the test pattern
    *---------------------------------------------------------------------*/
   for (dataWordIndex=0, subFrameIndex=0; subFrameIndex < 4; subFrameIndex++) {
      /* load the first word of each sub-frame with the respective sync word */
      xmitFrameData[dataWordIndex++] = (CEI_UINT16)syncWord[subFrameIndex];

      /* fill the remaining sub-frame with an incrementing value */
      while ( dataWordIndex < ((subFrameIndex+1) * subFrameSize) ) {
         xmitFrameData[dataWordIndex] = (CEI_UINT16)(dataWordIndex % 256);
         dataWordIndex++;
      }
   }

   if (status == ARS_NORMAL) {
      /*------------------------------------------------------------------*
       * since the transmit FIFO will support the entire frame, a single
       * invocation is all that is required.
       *------------------------------------------------------------------*/
TXSECONDFRAME:
      numberWordsToTransmit = frameSize;
      numberWordsToReceive = frameSize;
            
      status = ar_put_573_frame ( board,
                                  numberWordsToTransmit,
                                  &numberWordsToTransmit, 
                                  xmitFrameData );
      
      if (status != ARS_NORMAL) {
         printf ("Error status from ar_put_573_frame(): %s\n", ar_get_error(status) );
         (*errorCount)++;
      }
   
   
      /*------------------------------------------------------------------*
       * since the baud rate is 768bps, delay 3 seconds to allow 3 of the  
       * 4 subframes to be transmitted, then retrieve the entire frame.
       *------------------------------------------------------------------*/
      ar_sleep (3000);

      status = ar_get_573_frame ( board, 
                                  &numberWordsToReceive,
                                  recvFrameData );
   
      if (status == ARS_NODATA) {
         /* if the receiver didn't sync with the first frame transmission,
          * try transmitting one more frame first, if that doesn't work,
          * flush the receiver and try it all over again.
          */
         attemptsToWrapFrame++;
         if (attemptsToWrapFrame > 3) {
            printf ( "\nData Timeout waiting for ARINC 717 Reception\n");
            (*errorCount)++;
            status = ARS_WRAP_DATA_FAIL;
         }
         else if (attemptsToWrapFrame == 1) {
            goto TXSECONDFRAME;
         }
         else {
            status = ar_reset (board);
            ar_sleep (1000);
            goto RETRY573;
         }
      }
      
      /*------------------------------------------------------------------*
       * verify the received frame data matches the transmitted frame
       *------------------------------------------------------------------*/
      for ( dataWordIndex = 0; (dataWordIndex < frameSize) && (status == ARS_GOTDATA); dataWordIndex++ ) {
               
         if ( (recvFrameData[dataWordIndex] & CEIDEV_573_DATA_MASK) != xmitFrameData[dataWordIndex]) {
            printf ( "\nData Mismatch: Transmitted %4x Received: %4x \n",
                     (CEI_NATIVE_INT)xmitFrameData[dataWordIndex], 
                     (CEI_NATIVE_INT)(recvFrameData[dataWordIndex] & CEIDEV_573_DATA_MASK) );
            (*errorCount)++;
            status = ARS_WRAP_DATA_FAIL;
         }
      }
   
      if (status == ARS_GOTDATA) {
         printf ("Received a full ARINC 717 frame, Success!"); 
         status = ARS_NORMAL;
      }
   }

   printf ( "\n" );
   ar_stop (board);

   return status;
}



/*===========================================================================*
 * ENTRY POINT:   C U S T O M _ I N T E R R U P T _ H A N D L E R
 *===========================================================================*
 *
 * FUNCTION:    Example application-supplied interrupt service routine
 *
 * PARAMETERS:  deviceIndex   -- (input) device number to reference (0-15)
 *
 * DESCRIPTION: This routine provides the process by which the interrupt
 *              queue is handled when hardware interrupts are enabled.
 *              When invoked, this routine will request up to nine entries
 *              from the device interrupt queue and store the entries 
 *              in a local interrupt storage buffer.
 *
 *===========================================================================*/
#if defined (_WIN32)
   void _stdcall custom_interrupt_handler ( CEI_INT32 deviceIndex )
#elif defined (_LINUX_X86_) || defined (_LINUX_PPC_)
   void custom_interrupt_handler ( CEI_INT32 deviceIndex )
#else
   void custom_interrupt_handler ( CEI_INT32 deviceIndex, pCEI_UINT32 data )
#endif
{
   CEI_UINT32 interruptCount;        /* number of interrupt entries to fetch */
   CEI_INT32  status = ARS_NORMAL;              /* local API routine status  */

   /* Process only nine interrupt queue entries at a time */
   /* don't exceed the buffer size */
   if (globalInterruptCounter < (INTERRUPT_BUFFER_SIZE-9)) {
      interruptCount = 9;
      /* log any events from the h/w interrupt queue */
      status = ar_interrupt_queue_read ((CDEV_BOARD_TYPE)deviceIndex, 
                                        &interruptCount, 
                                        &globalQueueEntries[globalInterruptCounter]);
      if (status == ARS_GOTDATA) {
         globalInterruptCounter += interruptCount;
      }
   }
   return;
}


/*============================================================================*
 * ENTRY POINT:         D E M O _ P C I _ I N T E R R U P T S
 *============================================================================*
 *
 * FUNCTION:    Demonstrate the Receive Label Filter generated PCI Interrupt
 *
 * DESCRIPTION: This is a demonstration of the following CEI-x30 features:
 *              
 *                 Label Filter Interrupt Trigger Setup
 *                 CEI-x30 Default ISR execution
 *                 Application retrieval of API Interrupt Buffer Events
 *
 *============================================================================*/
#define TEST_MSG_1 0x55555555
#define TEST_MSG_2 0x77777777
CDEV_API_RET_TYPE demo_pci_interrupts ( CDEV_BOARD_TYPE board )
{
   CEI_UINT32 dataIndex;
   CDEV_CHAN_TYPE channel;        /* Current transmit channel being tested    */
   AR_SCHEDULED_MSG_ENTRY_TYPE msgEntry[2];
   CDEV_API_RET_TYPE status = ARS_NORMAL;
#ifdef DEFAULT_ISR
   CEI_UINT32 intQueueEntries[INTERRUPT_BUFFER_SIZE]; /* array to store intQ entries */
   CEI_UINT32 interruptCount;                  /* number of interrupts logged */
   printf ( "\n\nNow Executing the Interrupt Demonstration using the default ISR.\n\n");
#else
   CEI_INT32 isrStatus = ARS_NORMAL;
   globalInterruptCounter = 0;
   printf ( "\n\nNow Executing the Interrupt Demonstration using a custom ISR.\n\n");
#endif
   /*------------------------------------------------------------------------*
    *  Set the transmitter channels to the 100K bit rate & ODD parity.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < 2) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, ARU_SPEED_HIGH );
      status = ar_set_device_config ( board, channel, ARU_TX_PARITY, ARU_PARITY_ODD ); 
      status = ar_set_device_config ( board, channel, ARU_TX_DISABLE, disable );
      status = ar_set_device_config ( board, channel, ARU_TX_FIFO_ENABLE, AR_ON );
   }
   
   /*------------------------------------------------------------------------*
    *  Set the receiver channels to the 100K bit rate & ODD parity,
    *  also enable internal wrap.
    *------------------------------------------------------------------------*/
   for (channel = 0; (channel < 2) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_RECV_MODE, wrapMode );
      status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, ARU_SPEED_HIGH );
      status = ar_set_device_config ( board, channel, ARU_RX_PARITY, ARU_PARITY_ODD ); 
      status = ar_set_device_config ( board, channel, ARU_RX_MERGED_MODE, AR_OFF );
      status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_ON );
   }
   /* Setup the Receive Channel 0 and 1 Label Filter Tables to generate an
    * interrupt event for any received message content with label values of
    * 0x55 and 0x77 (see defines for TEST_MSG_* above).
    */
   status = (CEI_INT16)ar_enh_label_filter (board, 
                                            0, 
                                            (CEI_UINT16)( TEST_MSG_1 & 0xff), 
                                            ARU_ALL_SDI,
                                            ARU_ALL_ESSM,
                                            FILTER_INTERRUPT);
   status = (CEI_INT16)ar_enh_label_filter (board, 
                                            1, 
                                            (CEI_UINT16)( TEST_MSG_2 & 0xff), 
                                            ARU_ALL_SDI,
                                            ARU_ALL_ESSM,
                                            FILTER_INTERRUPT);
   
   /* Now setup the message scheduler with one entry for each of the test 
    * messages to be transmitted one each on transmitters 0 and 1, with a 
    * transmit count of 10 at a rate of 200ms (channel 1 offset by 100ms).
    */
   msgEntry[0].messageIndex = 0;
   msgEntry[0].board        = board;
   msgEntry[0].channel      = 0;
   msgEntry[0].rate         = 200;
   msgEntry[0].start        = 0;
   msgEntry[0].txCount      = 10;
   msgEntry[0].data         = TEST_MSG_1;

   msgEntry[1].messageIndex = 0;
   msgEntry[1].board        = board;
   msgEntry[1].channel      = 1;
   msgEntry[1].rate         = 200;
   msgEntry[1].start        = 100;
   msgEntry[1].txCount      = 10;
   msgEntry[1].data         = TEST_MSG_2;

   status = ar_define_msg_block (2, msgEntry);
   if (status != ARS_NORMAL) {
      printf ("ERROR status from ar_define_msg_block(): %s\n", ar_get_error(status) );
   }
   
   /* if this source was built to demonstrate a custom ISR, request installation of
    * the example ISR so it will be referenced during the hardware interrupt enable
    * processing.
    */
#ifndef DEFAULT_ISR
   isrStatus = ar_set_isr_function ( board, custom_interrupt_handler);
#endif
   /* enable PCI interrupts on the device */
   status = ar_set_device_config ( board, 0, ARU_HW_INTERRUPT_ENABLE, AR_ON );
   ar_sleep (10);

   /* enable data processing on the device, delay long enough for all 10 message
    * transmissions to occur on each channel, then terminate data processing.
    */
   status = ar_go (board);
   status = ar_set_device_config ( board, 0, ARU_INSERT_INT_Q_ENTRY, 0 );
   status = ar_set_device_config ( board, 0, ARU_INSERT_INT_Q_ENTRY, 0 );
   ar_sleep (2200);
   status = ar_stop (board);

#ifndef DEFAULT_ISR
   /* if this source was built to demonstrate a custom ISR, display the entries
    * captured and saved by the custom ISR.
    */
   if (globalInterruptCounter > 0) {
      printf("Custom ISR Processed %d Entries: ", (CEI_NATIVE_INT)globalInterruptCounter);
      for (dataIndex = 0; dataIndex < globalInterruptCounter; dataIndex++) {
         printf("%3d ", (int)globalQueueEntries[dataIndex]);
      }
   }
   else {
      printf ("Detected %d hardware interrupts", globalInterruptCounter);
   }
#else
   /* if this source was built to demonstrate the default API ISR, display the entries
    * captured and saved by the API.
    */
   interruptCount = INTERRUPT_BUFFER_SIZE;
   status = (CEI_INT16)ar_hw_interrupt_buffer_read(board, &interruptCount, intQueueEntries);
   if ((status == ARS_GOTDATA) && (interruptCount > 0)) {
      printf("\nAPI ISR Processed Entries:   %2d :: ", (CEI_NATIVE_INT)interruptCount);
      for (dataIndex = 0; dataIndex < interruptCount; dataIndex++) {
         printf("%3d ", (int)intQueueEntries[dataIndex]);
      }
   }
   else {
      printf ("Did not detect any hardware interrupts");
   }
#endif
   
   printf ("\n");
   /* disable PCI interrupts on the device */
   status = ar_set_device_config ( board, 0, ARU_HW_INTERRUPT_ENABLE, AR_OFF );
   
   return status;
}

/*===========================================================================*
 * ENTRY POINT:   D E M O _ P R O G R A M M A B L E _ C H A N N E L S 
 *===========================================================================*
 *
 * FUNCTION:    Demonstrate how to configure programmable channels
 *
 * DESCRIPTION: This is a demonstration routine that configures boards with
 *              programmable/shared transmit receive channels.  There is an
 *              assumption with this routine that programmable/shared channels
 *              occupy the upper receive channels when the receive/transmit
 *              channel count is unbalanced/unequal.
 *
 *===========================================================================*/
CEI_VOID demo_programmable_channels ( CDEV_BOARD_TYPE board )
{
   CDEV_CHAN_TYPE channel;                   /* channel index                      */
   CDEV_CHAN_TYPE firstProgrammableRxChannel;/* first programmable receive channel */
   CDEV_CHAN_TYPE firstTransmitChannel;      /* programmable channel index         */
   CDEV_API_RET_TYPE status = ARS_NORMAL;    /* utility status return value        */

   /* Determine the first programmable/shared receive channel, assuming all 
    * transmit channels are configured as programmable/shared.
    * NOTE: this assumption may not always be true for every board configuration.
    */
   firstProgrammableRxChannel = (CEI_INT16)((numTxChan < numRxChan) ? numRxChan-numTxChan : 0);
   
   /*------------------------------------------------------------------------*
    *  For the first demonstration, configure all programmable channels to be
    *  transmitters, and disable reception on the respective shared pins.
    *
    *  Set the channels to the 100K bit rate & ODD parity, external I/O.
    *------------------------------------------------------------------------*/
   /* first configure all fixed receive channels (non-programmable) for external operation */
   for (channel = 0; (channel < firstProgrammableRxChannel) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_RECV_MODE, AR_WRAP_OFF );
      status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, ARU_SPEED_HIGH );
      status = ar_set_device_config ( board, channel, ARU_RX_PARITY, AR_ON );
      status = ar_set_device_config ( board, channel, ARU_RX_MERGED_MODE, AR_OFF );
      status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_ON );
   }
   /* next disable all programmable/shared receive channels */
   for (channel = firstProgrammableRxChannel; (channel < numRxChan) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_OFF );   /* disable this receive buffer */
   }
   
   /* and last configure all programmable/shared transmit channels for external operation */
   for (channel = 0; (channel < numTxChan) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, ARU_SPEED_LOW );
      status = ar_set_device_config ( board, channel, ARU_TX_PARITY, ARU_PARITY_ODD ); 
      status = ar_set_device_config ( board, channel, ARU_TX_DISABLE, AR_OFF );       /* enable this transmitter circuit */
      status = ar_set_device_config ( board, channel, ARU_TX_FIFO_ENABLE, AR_ON );    /* enable this transmit buffer */
   }
   
   /*------------------------------------------------------------------------*
    *  For the second demonstration, configure only the upper half of the  
    *  programmable channels to be transmitters, and disable reception on 
    *  the respective shared pins.
    *
    *  Set the channels to the 100K bit rate & ODD parity, external I/O.
    *------------------------------------------------------------------------*/
   firstTransmitChannel = (CEI_INT16)(numTxChan / (CEI_INT16)2);
   firstProgrammableRxChannel = (CEI_INT16)((numTxChan < numRxChan) ? numRxChan-numTxChan : 0);
   
   /* first configure all fixed receivers for external operation, including 
    * configuring the first 1/2 programmable channels as receivers.
    */
   for (channel = 0; (channel < (firstProgrammableRxChannel+firstTransmitChannel)) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_RECV_MODE, AR_WRAP_OFF );
      status = ar_set_device_config ( board, channel, ARU_RX_BITRATE, ARU_SPEED_HIGH );
      status = ar_set_device_config ( board, channel, ARU_RX_PARITY, AR_ON );
      status = ar_set_device_config ( board, channel, ARU_RX_MERGED_MODE, AR_OFF );
      status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_ON );
   }
   /* next disable reception on the upper programmable/shared receive channels */
   for (channel = (CEI_INT16)(firstProgrammableRxChannel+firstTransmitChannel); (channel < numRxChan) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_RX_FIFO_ENABLE, AR_OFF );   /* disable this receive buffer */
   }
   
   /* and last configure the upper half of the programmable/shared channels as transmitters for external operation */
   for (channel = firstTransmitChannel; (channel < numTxChan) && (status == ARS_NORMAL); channel++) {
      status = ar_set_device_config ( board, channel, ARU_TX_BITRATE, ARU_SPEED_LOW );
      status = ar_set_device_config ( board, channel, ARU_TX_PARITY, ARU_PARITY_ODD ); 
      status = ar_set_device_config ( board, channel, ARU_TX_DISABLE, AR_OFF );       /* enable this transmitter circuit */
      status = ar_set_device_config ( board, channel, ARU_TX_FIFO_ENABLE, AR_ON );    /* enable this transmit buffer */
   }
   
}


/*=================================================================================*
 * ENTRY POINT: C H E C K _ E N H A N C E D _ F I R M W A R E _ C A P A B I L I T Y
 *=================================================================================*
 *
 * FUNCTION:  Process CEI-x30 BAR2 PROM Programming for Required Firmware Interface
 *
 * DESCRIPTION: The original CEI-x30 Firmware Interface only required 4Kb of PCI
 *              memory space - the size of BAR2.  The Enhanced CEI-x30 Firmware
 *              Interface requires 512Kb, so the size of BAR2 on any CEI-x30
 *              device must be programmed to 512Kb.  This routine handles the
 *              operator prompts and routine invocations to program BAR2 to either
 *              512Kb or restore it to 4Kb.
 *
 *=================================================================================*/
CEI_VOID check_enhanced_firmware_capability (CDEV_BOARD_TYPE board, CEI_NATIVE_INT restoreConfig)
{
   CDEV_API_RET_TYPE status;
   CDEV_PARM_SSI_TYPE no_care = 0;          /* placeholder parameter */
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
   CEI_CHAR updateSelection;        /* EEPROM update request selection   */
   CEI_NATIVE_INT result;           /* status from scanf operator prompt */
#endif
   /* If the user requested this board be restored to the original x30 4Kb BAR2 size, do it. */
   if (restoreConfig == 1) {
      printf ("\nYour device is being updated to the original CEI-x30 4Kb Interface.\n");
      printf (  "DO NOT REMOVE POWER OR INTERRUPT THIS PROGRAM while this update is executing.\n");
      status = ar_set_device_config ( board, 0, ARU_HW_ENHANCE_UPDATE, AR_OFF );
      if (status == ARS_NORMAL) {
         printf ("\nThe device has been successfully reprogrammed.\n");
         printf (  "You must now reboot/restart the host computer in order for the  \n");
         printf (  "CEI-x30 Interface resource update to be recognized by the BIOS. \n");
      }
      else {
         printf ("\nThe EEPROM program utility failed to update the specified CEI-x30 device.\n");
         printf (  "Please reboot/restart your host computer and try again.                  \n");
      }
      exit (0);
   }

   /*------------------------------------------------------------------------*
    *  Open a session with the device to check for compatibility with the
    *  CEI-x30 Enhanced API for which this distribution only supports.
    *  If the board is defined to have the smaller BAR2 memory region,
    *  report that to the operator and if on a supported o/s, allow the
    *  option to reprogram the board and reboot.
    *------------------------------------------------------------------------*/
   status = ar_get_device_config ( board, 0, ARU_HW_ENHANCE_CHECK, &no_care );
   if (status == ARS_HW_CONSISTENCY) {
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
      printf ("\n----------------------------------------------------------------------------\n");
      printf (  "                   *** HARDWARE UPDATE NOTIFICATION ***                     \n");
      printf (  "----------------------------------------------------------------------------\n");
      printf ("\nThe specified CEI-x30 device is programmed for the previous standard CEI-x30\n");
      printf (  "PCI interface.  The CEI-x30 Enhanced Interface supported by this and all    \n");
      printf (  "subsequent API releases requires a modified PCI Interface definition that   \n");
      printf (  "can be reprogrammed through the API and completed with a host restart.      \n");
      printf (  "Would you like to reprogram this device to support the new Enhanced CEI-x30 \n");
      printf (  "Interface? (respond with 'y' for yes or 'n' for no): ");
      result = scanf("%c", &updateSelection);
      if ((updateSelection == 'y') || (updateSelection == 'Y')) {
         printf ("\nYour device is being updated to support the CEI-x30 Enhanced Interface.      \n");
         printf (  "DO NOT REMOVE POWER OR INTERRUPT THIS PROGRAM while this update is executing.\n");
         status = ar_set_device_config ( board, 0, ARU_HW_ENHANCE_UPDATE, AR_ON );
         if (status == ARS_NORMAL) {
            printf ("\nThe device has been successfully reprogrammed.\n");
            printf (  "You must now reboot/restart the host computer in order for the  \n");
            printf (  "CEI-x30 Interface resource update to be recognized by the BIOS. \n");
         }
         else {
            printf ("\nThe EEPROM program utility failed to update the specified CEI-x30 device.\n");
            printf (  "Please reboot/restart your host computer and try again.                  \n");
         }
      }
      else {
         printf ("\nYou have selected not to update your CEI-x30 device for use with the Enhanced \n");
         printf (  "CEI-x30 PCI Interface.  Since this distribution does not support the previous \n");
         printf (  "Standard CEI-x30 PCI Interface, you cannot proceed to communicate with this   \n");
         printf (  "device using this API/distribution.  See the CEI-x30-SW User Manual for a     \n");
         printf (  "detailed description of the CEI-x30 Enhanced PCI Interface, and the benefits  \n");
         printf (  "provided by updating your CEI-x30 device and incorporating this enhancement   \n");
         printf (  "into your ARINC application; otherwise, instructions for altering the software\n");
         printf (  "distribution files for use with the standard CEI-x30 API are provided.        \n");
      }
      printf ("Press any key to exit\n");
      getch();
      exit (0);
#else
   #if defined (AUTO_EMBEDDED_UPDATE)
      printf ("\nYour device is being updated to support the CEI-x30 Enhanced Interface.      \n");
      printf (  "DO NOT REMOVE POWER OR INTERRUPT THIS PROGRAM while this update is executing.\n");
      status = ar_set_device_config ( board, 0, ARU_HW_ENHANCE_UPDATE, AR_ON );
      if (status == ARS_NORMAL) {
         printf ("\nThe device has been successfully reprogrammed.                          \n");
         printf (  "You must now reboot/restart the host computer in order for the Enhanced \n");
         printf (  "CEI-x30 Interface resource update to be recognized by the BIOS.         \n");
      }
      else {
         printf ("\nThe device reprogramming failed miserably.                         \n");
         printf (  "You may now reboot/restart the host computer; however, the Enhanced\n");
         printf (  "CEI-x30 Interface resource update may not be operational.          \n");
      }
   #else
      printf ("\nThe specified CEI-x30 device is programmed for the previous standard CEI-x30  \n");
      printf (  "PCI interface.  The CEI-x30 Enhanced Interface supported by this and all      \n");
      printf (  "subsequent API releases requires a modified PCI Interface definition that can \n");
      printf (  "be reprogrammed through the API and completed with a host restart.  If you    \n");
      printf (  "would like to reprogram this device to support the Enhanced CEI-x30 Interface,\n");
      printf (  "uncomment the #define AUTO_EMBEDDED_UPDATE at the top of this file, rebuild,  \n");
      printf (  "and run again.  See the instructions for your o/s in the CEI-x30-SW User Manual.\n");
   #endif
   exit (0);
#endif
   }
}



/*============================================================================*
 * ENTRY POINT:             E R R O R _ E X I T
 *============================================================================*
 *
 * FUNCTION:    Print out an error message and exit.
 *
 * DESCRIPTION: Call the ARINC error message utility, print out the message,
 *      and exit.
 *
 *============================================================================*/

CEI_VOID error_exit (CDEV_PARM_SSI_TYPE status)
{
   printf ("\n\nFATAL ERROR: BOARD FAILED INITIALIZATION!\n");
   printf ("Error reported: \'%s\'\n", ar_get_error(status));
   if (status == ARS_FW_NOT_SUPPORTED) {
      printf ("This version of CEI-x30-SW requires your board be programmed with\n");
      printf ("firmware version 4.13 or later.  Contact Abaco Systems Avionics Support\n");
      printf ("for assistance in upgrading your board to work with this distribution.\n");
   }
#if !defined (VXW_PCI_X86) && !defined (VXW_PCI_PPC) && !defined (_LINUX_X86_) && !defined (_LINUX_PPC_) && !defined (INTEGRITY_PCI_PPC) && !defined (INTEGRITY_PCI_X86)
   printf ("Press any key to exit\n");
   getch();
#endif
   exit (0);
}

